package coms6998_11.cs.columbia.edu.calculator;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.Stack;

public class CalculatorBrain {

	/**
	 * Object version for serialization.
	 */
	
	// Number of characters to preallocate when converting a stack value into a 
	// string or storing input from user.
	private static final int TYPICAL_LENGTH = 32;
	// 4x the above
	private static final int TYPICAL_LENGTH_X4 = 128;
	
	// How many digits of precision (decimal places) are used internally in 
	// calculations.
	private static final int INTERNAL_SCALE = 32;

	private final Stack<BigDecimal> stack;

  // Initial scale is 2 decimal places, as that's the most useful for general 
	// everyday calculations.
	private int scale = 2;

	public CalculatorBrain() {
		super();
		this.stack = new Stack<BigDecimal>();
	}

	/**
	 * Pushes a value onto the stack.
	 * @param number A valid decimal number, in a String. Usually taken from the 
	 * InputBuffer.
	 */
	public void push(final String number) {
		final BigDecimal newnum = new BigDecimal(number);
		this.stack.push(newnum);
	}
	
	/**
	 * Returns whether the stack is empty.
	 */
	public boolean isEmpty() {
		return this.stack.isEmpty();
	}
	
	/**
	 * Gets the contents of the stack as a string.
	 * @param levels the number of levels of stack to return
	 * @return a text representation of the stack
	 */
	public StringBuilder toString(final int levels) {
	  final StringBuilder result = new StringBuilder(TYPICAL_LENGTH_X4);
	  if (this.stack != null) {
	    final int depth = this.stack.size();
	    for (int i = 0; i < levels; i++) {
	      if (i != 0) {
	        result.append('\n');
	      }
	      final int idx = depth - levels + i;
	      if (idx >= 0) {
	        result.append(formatNumber(this.stack.get(idx)));
	      }
	    }
	  }
	  return result;
	}
	
	/**
	 * Formats a BigDecimal number to a fixed number of decimal places, and adds 
	 * thousands commas.
	 * @param number
	 * @return
	 */
	private String formatNumber(final BigDecimal number) {
		final StringBuilder result = new StringBuilder(TYPICAL_LENGTH);
		result.append(number.setScale(this.scale, 
		    RoundingMode.HALF_UP).toPlainString());
		if (this.scale > 0) {
			if (result.indexOf(".") == -1) {
				result.append('.');
			}
			final int zerosAfterPoint = result.length() - result.indexOf(".") - 1;
			for (int i = zerosAfterPoint; i < this.scale; i++) {
				result.append('0');
			}
		}
		// Add commas
		int dot = result.indexOf(".");
		if (dot < 1) {
			dot = result.length();
		}
		int lowindex = 0;
		if (result.charAt(0) == '-') {
			lowindex = 1;
		}
		for (int i = dot - 3; i > lowindex; i -= 3) {
			result.insert(i, ',');
		}
		return result.toString();
	}
	
	/**
	 * Drops the top element from the stack.
	 */
	public void drop() {
		if (!this.stack.isEmpty()) {
			this.stack.pop();
		}
	}
	
	/**
	 * Adds together the top two elements on the stack, and replaces them with
	 * the result.
	 */
	public void add() {
		if (this.stack.size() > 1) {
			final BigDecimal x = this.stack.pop();
			final BigDecimal y = this.stack.pop();
			final BigDecimal r = y.add(x);
			this.stack.push(r);
		}
	}
	
	/**
	 * Subtracts the top number on the stack from the number beneath it, and 
	 * replaces them both with the result.
	 */
	public void subtract() {
		if (this.stack.size() > 1) {
			BigDecimal x = this.stack.pop();
			BigDecimal y = this.stack.pop();
			BigDecimal r = y.subtract(x);
			this.stack.push(r);
		}
	}
	
	/**
	 * Multiplies the top two numbers on the stack together, and replaces them 
	 * with the result.
	 */
	public void multiply() {
		if (this.stack.size() > 1) {
			BigDecimal x = this.stack.pop();
			BigDecimal y = this.stack.pop();
			BigDecimal r = y.multiply(x);
			this.stack.push(r);
		}
	}
	
	
	/**
	 * Uses the top number on the stack to divide the number beneath it.
	 * Replaces both with the result of the division.
	 * e.g. before: x y  after: x/y   before:  4 2  after: 2
   * @return an error message, or null if there is no error
	 */
	public String divide() {
		String result = null;
		if (this.stack.size() > 1) {
			BigDecimal x = this.stack.pop();
			BigDecimal y = this.stack.pop();
			// We use HALF_EVEN rounding because this statistically minimizes 
			// cumulative error during repeated calculations.
			try {
				BigDecimal r = y.divide(x, INTERNAL_SCALE,
						RoundingMode.HALF_EVEN);
				this.stack.push(r);
			} catch (ArithmeticException e) {
				result = e.getMessage();
			}
		}
		return result;
	}
	public void setScale(final int newscale) {
		this.scale = newscale;
	}
	
	public void setScale() {
		if (!this.stack.isEmpty()) {
			BigDecimal x = this.stack.pop();
			int sc = x.intValue();
			if (sc < INTERNAL_SCALE) {
				setScale(sc);
			}
		}
	}
	/**
	 * Duplicates the top element on the stack.
	 */
	public void dup() {
		if (!this.stack.isEmpty()) {
			final BigDecimal topnum = this.stack.peek();
			this.stack.push(topnum);
		}
	}
	/**
	 * Gets the current display scale.
	 * @return
	 */
	public int getScale() {
		return this.scale;
	}
	
	public void emptyStack(){
		while(!this.stack.isEmpty())
			this.stack.pop();
	}
	
	public void performOperation(char c){
        switch (c) {
        case '+':
          add();
          break;
        case '-':
          subtract();
          break;
        case '*':
          multiply();
          break;
        case '/':
          divide();
          break;
        default:
          break;	  
        }
	}
}
